r"""generate Advanced SubStation Alpha(ASS) subtitles from meta data.

Typical usage:
    danmaku = Danmaku()
    meta_data = {
        'start': '0:0:0.23',
        'end': '0:0:0.45',
        'mode': 'top',
        'content': 'lol',
        'fontsize': r'\fs23',
        'color': r'\c&HFFFFFF&'
    }
    dia_log = danmaku.genera_dialog(**meta_data)
    danmaku.add(dia_log)
    danmaku.output_subtitle()
"""
from typing import Optional
import os
import random


class Danmaku(object):
    """Danmaku."""
    def __init__(self, ass_file_path: Optional[str] = None):
        """read ass header from disk."""
        if not ass_file_path:
            ass_file_path = os.path.join(
                os.path.dirname(__file__), 'resource', 'ass_header.txt')

        with open(ass_file_path, 'r', encoding='utf-8') as f:
            self.ass_header = f.read()

        self.screen_width = 560
        self.screen_height = 420

        self.subtitles = []

    def edit_header(self, title: str,
                    width: Optional[str] = None,
                    height: Optional[str] = None) -> None:
        """edit ass header according to title and screen's resolution."""
        if not width:
            width = self.screen_width
        else:
            self.screen_width = width

        if not height:
            height = self.screen_height
        else:
            self.screen_height = height

        self.ass_header = self.ass_header.replace('@title@', title)
        self.ass_header = self.ass_header.replace('@width@', str(width))
        self.ass_header = self.ass_header.replace('@height@', str(height))

        self.subtitles.append(self.ass_header)

    # TODO: advanced algorithm
    def generate_dialog(self, *, start: str, end: str, mode: str, content: str,
                        fontsize: Optional[str] = '',
                        color: Optional[str] = '',
                        ) -> str:
        """generate dialog with meta data."""
        height = random.randint(0, self.screen_height)  # need to be improved
        content_len = 12 * len(content)

        if mode == 'normal':
            move = (r'\an7\move('
                    f'{self.screen_width}, {height},'
                    f' {-content_len}, {height})')
        elif mode == 'reverse':
            move = (r'\an9\move('
                    f'0, {height}, '
                    f'{content_len + self.screen_width}, {height})')
        elif mode == 'bottom':
            move = r'\an2\pos(' + f'{self.screen_width/2}, {height})'
        elif mode == 'top':
            move = r'\an8\pos(' + f'{self.screen_width/2}, {height})'
        else:
            raise NotImplementedError

        code = f'{{{fontsize}{move}{color}}}'
        return f'Dialogue: 0,{start},{end},Danmaku,,0,0,0,,{code}{content}'

    def add_dialog(self, dialog: str) -> None:
        """add a dialogue into subtitles."""
        self.subtitles.append(dialog)

    def output_subtitle(self) -> str:
        """return subtitle."""
        return '\n'.join(self.subtitles)



"""convert json subtitles to ass subtitles."""
from video_dl.danmaku import Danmaku


class Convertor(object):
    """convert json to ass."""
    def __init__(self, file_path: str = None):
        self.danmaku = Danmaku(file_path)
        self.screen_width = 560  # width of screen
        self.screen_height = 420  # height of screen
        self.move_time = 8  # duration time of move subtitle
        self.fixed_time = 4  # duration time of fixed subtitle

        self.result = []

    def edit_header(self, title: str) -> None:
        self.danmaku.edit_header(title)

    def ms2datetime(self, ms: int) -> str:
        """convert ms to datetime."""
        hour = int(ms/(1000*60*60))
        minute = int(ms/(1000*60)) % 60
        second = int(ms/1000) % 60
        ms = ms % 1000

        return f'{hour}:{minute}:{second}.{ms}'

    def json2ass(self, json_data: dict) -> None:
        text = json_data['content']

        if json_data['mode'] in (1, 2, 3, 7, 8, 9):
            duration = self.move_time * 1000
            mode = 'normal'
        elif json_data['mode'] == 6:
            duration = self.move_time * 1000
            mode = 'reverse'
        elif json_data['mode'] == 4:  # bottom
            duration = self.fixed_time * 1000
            mode = 'bottom'
        elif json_data['mode'] == 5:  # top
            duration = self.fixed_time * 1000
            mode = 'top'

        start = self.ms2datetime(json_data['progress'])
        end = self.ms2datetime(json_data['progress'] + duration)
        color = r'\c&H' + str(hex(json_data['color']))[-6:] + '&'

        self.danmaku.add_dialog(self.danmaku.generate_dialog(**{
            'start': start,
            'end': end,
            'mode': mode,
            'content': text,
            'color': color,
        }))

    def output(self) -> str:
        return self.danmaku.output_subtitle()



"""Spider for ixigua.com"""
from urllib.parse import urlparse, urljoin

from video_dl.extractor import Extractor
from video_dl.spider import Spider
from video_dl.toolbox import info
from video_dl.video import Media


class IXiGuaSpider(Spider):
    """spider for ixigua.com"""
    site = 'ixigua.com'
    home_url = 'https://www.ixigua.com'

    def __init__(self):
        super().__init__()

        self.cookies = None

    async def before_download(self) -> None:
        await self.parse_html(self.url)

    async def parse_html(self, target_url: str) -> None:
        """extract key information from html source code.

        Args:
            target_url: target url copied from online vide website.
        """
        info('url', target_url)
        target_url = urljoin(target_url, '?wid_try=1')
        extractor = Extractor.create(target_url)

        # Anti-anti-crawler
        if not self.cookies:
            await self.fetch_html(target_url)
            result = urlparse(target_url)
            meta_data = {
                'pathname': result.path,
                'href': target_url,
                'search': f'?{result.query}' if result.query else '',
                'referrer': target_url,
                'ac_nonce': self.session.cookie_jar.filter_cookies(target_url)['__ac_nonce'].value,
            }
            self.cookies = extractor.get_cookies(meta_data)

        resp, _ = await self.fetch_html(target_url, cookies=self.cookies)
        video = self.create_video()
        video.title = extractor.get_title(resp)

        for mp4 in extractor.get_mp4_video_url(resp):
            video.add_media(Media(**mp4))
        self.video_list.append(video)

        if self.lists:
            info('list', 'not implemented yet!')


"""Extract information from html source code.

Typical usage:
    url = 'https://www.bilibili.com/video/BV15L411p7M8'
    extractor = Extractor.create(url)  # will return a BilibiliVideoExtractor
"""


class Extractor(object):
    """Base class of Extractors.

    subclass of Extractor should provide a public attribute:
        pattern: Extractor will use this list to create a specific Spider for target url.
         For example, BilibiliVideoExtractor's pattern is [re.compile('bilibili.com/video/BV.*')],
         will auto match to a url like 'https://www.bilibili.com/video/BV346'.
    """
    @classmethod
    def create(cls, url: str):
        """return a specific Extractor depends on url."""
        for subclass in cls.__subclasses__():
            for pattern in subclass.pattern:
                if pattern.search(url):
                    return subclass()
        raise NotImplementedError



"""extract information from html source code of pornhub.com."""
import json
import re

from video_dl.extractor import Extractor


class PornhubExtractor(Extractor):
    """pornhub information extractor."""
    pattern = [
        re.compile(r'pornhub.com/view_video.php\?viewkey='),
    ]

    # re patterns to extract information from html source code
    re_video_show = re.compile(r'VIDEO_SHOW.*?({.*?});')
    re_mp4_url = re.compile(r'player_mp4_seek.*?//.*?;(.*?);flashvars', re.S)
    re_key_value = re.compile('var (.*?)=(.*)')
    re_drop_char = re.compile('[ +"]')

    def get_title(self, resp: str) -> str:
        """get video's title from html source code."""
        video_show = json.loads(self.re_video_show.search(resp).group(1))
        return video_show['videoTitle']

    def get_mp4_video_url(self, resp: str) -> str:
        """pornhub save its mp4 links in this url."""
        url_string = self.re_mp4_url.search(resp).group(1)
        url_string = re.sub(r'/\*.*?\*/', '', url_string)  # drop comment
        url_string = re.sub(r'[\n\t]', '', url_string)  # drop extra space

        result = {}
        url = []
        for item in url_string.split(';'):
            key, value = self.re_key_value.search(item).groups()
            if key != 'media_0':  # get tamporary variables
                result[key] = self.re_drop_char.sub('', value)
            else:
                for k in value.split('+'):  # ger final url
                    url.append(result[k.strip()])
        return ''.join(url)



from .bilibili import (BilibiliVideoExtractor,
                       BilibiliBangumiExtractor,
                       BilibiliSpider)
from .pornhub import PornhubSpider, PornhubExtractor
from .ixigua import IXiGuaSpider, IXiGuaExtractor



"""extract information from html source code of ixigua.com."""
from typing import Optional
import base64
import json
import os
import re
import subprocess

from video_dl.extractor import Extractor


class IXiGuaExtractor(Extractor):
    """ixigua information extractor."""
    pattern = [
        re.compile(r'www.ixigua.com/\d+'),
    ]

    # re patterns to extract information from html source code
    re_video = re.compile(r'window\._SSR_HYDRATED_DATA=(.*?)</script>', re.S)

    def __init__(self, js_path:Optional[str] = None):
        super().__init__()

        if not js_path:
            js_path = os.path.join(os.path.dirname(__file__), '..', '..', 'resource', 'ixigua_acrawler.js')
        self.js_path = js_path

        with open(js_path, 'r', encoding='utf-8') as f:
            self.js_code = f.read()

    def get_cookies(self, meta_data: dict) -> dict:
        for key, value in meta_data.items():
            self.js_code = self.js_code.replace(f'@{key}@', value)
        js_path = os.path.splitext(self.js_path)[0] + '.tmp.js'

        with open(js_path, 'w', encoding='utf-8') as f:
            f.write(self.js_code)
        result = subprocess.run(['node', js_path], stdout=subprocess.PIPE, check=True)

        os.remove(js_path)
        return {
            '__ac_nonce': meta_data['ac_nonce'],
            '__ac_signature': result.stdout.decode('utf-8').strip(),
            '__ac_referer': '__ac_blank',
        }

    def get_title(self, resp: str) -> str:
        """get video's title from html source code."""
        json_string = self.re_video.search(resp).group(1)
        json_string = json_string.replace('undefined', 'null')
        video = json.loads(json_string)
        return video['anyVideo']['gidInformation']['packerData']['video']['title']

    def get_mp4_video_url(self, resp: str) -> list:
        """ixigua.com save its mp4 links in this dictionary."""
        json_string = self.re_video.search(resp).group(1)
        json_string = json_string.replace('undefined', 'null')
        video = json.loads(json_string)

        video_list = video['anyVideo']['gidInformation']['packerData']['video']['videoResource']['normal']['video_list']
        for item in video_list.values():
            yield {
                'url': base64.b64decode(item['main_url']).decode('utf-8'),
                'size': item['size'],
                'desc': item['definition']
            }


"""Spider for pornhub.com"""
from video_dl.extractor import Extractor
from video_dl.spider import Spider
from video_dl.toolbox import info
from video_dl.video import Media


class PornhubSpider(Spider):
    """spider for pornhub.com"""
    site = 'pornhub.com'
    home_url = 'https://www.pornhub.com'

    async def before_download(self) -> None:
        await self.parse_html(self.url)

    async def parse_html(self, target_url: str) -> None:
        """extract key information from html source code.

        Args:
            target_url: target url copied from online vide website.
        """
        info('url', target_url)
        resp, _ = await self.fetch_html(target_url)
        extractor = Extractor.create(target_url)

        video = self.create_video()
        video.title = extractor.get_title(resp)

        if self.lists:
            info('list', 'not implemented yet!')

        mp4_url = extractor.get_mp4_video_url(resp)
        mp4_dict = await self.fetch_json(mp4_url)

        for mp4 in mp4_dict:
            quality = mp4['quality']
            video.add_media(Media(**{
                'url': mp4['videoUrl'],
                'size': int(quality),
                'desc': f'{quality}P',
            }))

        self.video_list.append(video)
